{*******************************************************}
{                                                       }
{                                                       }
{       Unit : main                                     }
{       Created by Tsviatko Jongov                      }
{       http://tsviatko.jongov.com                      }
{       tsviatko@jongov.com                             }
{                                                       }
{       Date : 10.11.2006                               }
{                                                       }
{       DirectShow video edge detect filter.            }
{                                                       }
{*******************************************************}
unit main;


interface

uses
  Windows, SysUtils, BaseClass, ActiveX, DirectShow, DSUTil, Classes, Dialogs,
  EdgeDetectUtils, SyncObjs;

const
  CLSID_JongovEdgeDetectFilter: TGUID = '{7543E9D2-A28D-41DD-9E36-978B472ED1A5}';

  //Edge detection type
  edtNone       = 0;
  edtHorizontal = 1;
  edtVertical   = 2;

type
  IEdgeDetectControl = interface
  ['{3BB65567-4AB7-40FA-B943-F8AC0C0E6321}']
    function GetEdgeDetectType(out EdgeDetectType: Integer): HRESULT; stdcall;
    function SetEdgeDetectType(EdgeDetectType: Integer): HRESULT; stdcall;

    function GetCoefficients(out K13, K2: Double): HRESULT; stdcall;
    function SetCoefficients(K13, K2: Double): HRESULT; stdcall;
  end;

  TEdgeDetectFilter = class(TBCTransformFilter, IPersist, IEdgeDetectControl,
      ISpecifyPropertyPages)
    FThisInstance          : Integer;
    FMediaPos              : IMediaPosition;
    FGraph                 : IFilterGraph;

    FEdgeDetectType        : Integer;
    FK1_3                  : Double; //Coefficients 1 and 3
    FK2                    : Double; //Coefficient 2
    FCriticalSection       : TCriticalSection;

    FWidth                 : Integer;
    FHeight                : Integer;

    FEdgeDetector          : TEdgeDetector;
  private
    function CurrCopyMediaSample(Source, Dest: IMediaSample): HRESULT;
  public
    constructor Create(ObjName: string; unk: IUnKnown; out hr: HRESULT);
    constructor CreateFromFactory(Factory: TBCClassFactory; const Controller: IUnknown); override;
    destructor Destroy; override;

    function Transform(pIn, pOut: IMediaSample): HRESULT; overload; override;

    function DecideBufferSize(Alloc: IMemAllocator;
      Properties: PAllocator_Properties): HRESULT; override;

    function CheckTransform(mtIn, mtOut: PAM_Media_Type): HRESULT; override;{Buffer Size in seconds}

    function CheckInputType(mtIn: PAM_Media_Type): HRESULT; override;

    function GetMediaType(Position: Integer;
              out MediaType: pAM_Media_Type): HRESULT; override;

    function JoinFilterGraph(pGraph: IFilterGraph; pName: PWideChar): HRESULT; override; stdcall;

    function CompleteConnect(direction: TPin_Direction; ReceivePin: IPin): HRESULT; override;

    function Stop: HRESULT; override; stdcall;
    function Pause: HRESULT; override; stdcall;

    {ISpecifyPropertyPages}
    function GetPages(out pages: TCAGUID): HRESULT; stdcall;

    {IEdgeDetectControl}
    function GetEdgeDetectType(out EdgeDetectType: Integer): HRESULT; stdcall;
    function SetEdgeDetectType(EdgeDetectType: Integer): HRESULT; stdcall;

    function GetCoefficients(out K13, K2: Double): HRESULT; stdcall;
    function SetCoefficients(K13, K2: Double): HRESULT; stdcall;
  end;

implementation


uses
  PropEdit;

{---------------------------- TEdgeDetectFilter -------------------------}
function TEdgeDetectFilter.CheckInputType(mtIn: PAM_Media_Type): HRESULT;
begin
  Result := E_FAIL;
  try
    if IsEqualGUID(mtIn.majortype, MEDIATYPE_Video) and
       IsEqualGUID(mtIn.subtype, MEDIASUBTYPE_RGB32) and
       IsEqualGUID(mtIn.formattype, FORMAT_VideoInfo) and
       (mtIn.cbFormat >= SizeOf(TVideoInfoHeader)) then
    begin
      if (PVideoInfoHeader(mtIn.pbFormat)^.bmiHeader.biWidth > 0) and
         (PVideoInfoHeader(mtIn.pbFormat)^.bmiHeader.biHeight > 0) then
      begin
        Result := S_OK;
      end;
    end;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.CheckInputType call...');
  end;
end;

function TEdgeDetectFilter.CheckTransform(mtIn, mtOut: PAM_Media_Type): HRESULT;
var
  pInput  : PVideoInfoHeader;
  pOutput : PVideoInfoHeader;
begin
  Result := E_FAIL;
  try
    Result := CheckInputType(mtIn);
    if FAILED(Result) then
      Exit;

    Result := CheckInputType(mtOut);
    if FAILED(Result) then
      Exit;

    if not IsEqualGUID(mtOut.formattype, FORMAT_VideoInfo) then
    begin
      Result := E_INVALIDARG;
      Exit;
    end;

    if (mtIn.cbFormat < SizeOf(TVIDEOINFOHEADER)) or
       (mtOut.cbFormat < SizeOf(TVIDEOINFOHEADER)) then
    begin
      Result := E_INVALIDARG;
      Exit;
    end;

    pInput := PVideoInfoHeader(mtIn.pbFormat);
    pOutput := PVideoInfoHeader(mtOut.pbFormat);

    if CompareMem(@pInput.bmiHeader, @pOutput.bmiHeader, sizeof(BITMAPINFOHEADER)) then
      Result := NOERROR
    else
      Result := E_INVALIDARG;

    Result := S_OK;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.CheckTransform call...');
  end;
end;

constructor TEdgeDetectFilter.Create(ObjName: string; unk: IInterface; out Hr: HRESULT);
begin
  try
    inherited Create(ObjName, unk, CLSID_JongovEdgeDetectFilter);

    FEdgeDetectType := edtNone;
    FEdgeDetector := TEdgeDetector.Create;

    FK1_3 := -0.1;
    FK2 := 0.2;

    FCriticalSection := TCriticalSection.Create;
    IsMultiThread := True;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.Create call...');
  end;
end;

constructor TEdgeDetectFilter.CreateFromFactory(Factory: TBCClassFactory;
  const Controller: IInterface);
var
  hr: HRESULT;
begin
  try
    Create(Factory.Name, Controller, hr);
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.CreateFromFactory call...');
  end;
end;

destructor TEdgeDetectFilter.Destroy;
begin
  try
    FreeAndNil(FCriticalSection);
    FreeAndNil(FEdgeDetector);

    inherited;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.Destroy call...');
  end;
end;

function TEdgeDetectFilter.Transform(pIn, pOut: IMediaSample): HRESULT;
begin
  Result := E_FAIL;
  try
    CurrCopyMediaSample(pIn, pOut);
    Result := S_OK;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.Transform call...');
  end;
end;

function TEdgeDetectFilter.JoinFilterGraph(pGraph: IFilterGraph;
  pName: PWideChar): HRESULT;
begin
  Result := inherited JoinFilterGraph(pGraph, pName);
  if pGraph <> nil then
  begin
    FMediaPos := pGraph as IMediaPosition;
    FGraph    := pGraph;
  end
  else begin
    FMediaPos := nil;
    FGraph    := nil;
  end;
end;

function TEdgeDetectFilter.CompleteConnect(direction: TPin_Direction;
  ReceivePin: IPin): HRESULT;
var
  pmt: PAM_Media_Type;
begin
  Result := E_FAIL;
  try
    ASSERT(FInput <> nil);
    ASSERT(FOutput <> nil);

    // if we are not part of a graph, then don't indirect the pointer
    // this probably prevents use of the filter without a filtergraph
    if(FGraph = nil) then
    begin
      result := VFW_E_NOT_IN_GRAPH;
      exit;
    end;

    // Always reconnect the input to account for buffering changes
    //
    // Because we don't get to suggest a type on ReceiveConnection
    // we need another way of making sure the right type gets used.
    //
    // One way would be to have our EnumMediaTypes return our output
    // connection type first but more deterministic and simple is to
    // call ReconnectEx passing the type we want to reconnect with
    // via the base class ReconeectPin method.

    if(direction = PINDIR_OUTPUT) then
    begin
      if FInput.IsConnected then
      begin
        result := ReconnectPin(FInput, FOutput.AMMediaType);
        exit;
      end;
      result := NOERROR;
      exit;
    end;

    ASSERT(direction = PINDIR_INPUT);

    // Reconnect output if necessary

    if FOutput.IsConnected then
    begin
      pmt := FInput.CurrentMediaType.MediaType;
      if (not TBCMediaType(pmt).Equal(FOutput.CurrentMediaType.MediaType)) then
      begin
        result := ReconnectPin(FOutput, FInput.CurrentMediaType.MediaType);
        exit;
      end;
    end;
    Result := NOERROR;

    Result := inherited CompleteConnect(direction, ReceivePin);

    if (Result = NOERROR) and (direction = PINDIR_INPUT) then
    begin
//    CopyMediaType(@FCurrMediaType, FInput.CurrentMediaType.MediaType);

      FWidth := PVideoInfoHeader(FInput.CurrentMediaType.MediaType.pbFormat)^.bmiHeader.biWidth;
      FHeight := PVideoInfoHeader(FInput.CurrentMediaType.MediaType.pbFormat)^.bmiHeader.biHeight;

      if (FWidth <= 0) or (FHeight <= 0) then
      begin
        Result := E_FAIL;
        Exit;
      end;

    end;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.CompleteConnect call...');
  end;
end;

function TEdgeDetectFilter.Pause: HRESULT;
begin
  Result := inherited Pause;
end;

function TEdgeDetectFilter.Stop: HRESULT;
begin
  Result := inherited Stop;
end;

function TEdgeDetectFilter.DecideBufferSize(Alloc: IMemAllocator;
  Properties: PAllocator_Properties): HRESULT;
var
  InProps, Actual : TALLOCATOR_PROPERTIES;
  InAlloc         : IMemAllocator;
begin
  // Is the input pin connected
  if not FInput.IsConnected then
    begin
      Result := E_UNEXPECTED;
      Exit;
    end;

  ASSERT(Alloc <> nil);
  ASSERT(Properties <> nil);

  Properties.cBuffers := 1;
  Properties.cbAlign  := 1;


  // Get input pin's allocator size and use that
  Result := FInput.GetAllocator(InAlloc);
  if SUCCEEDED(Result) then
  begin
    Result := InAlloc.GetProperties(InProps);
    if SUCCEEDED(Result) then
      Properties.cbBuffer := InProps.cbBuffer;

    InAlloc := nil;
  end;

  if FAILED(result) then Exit;

  ASSERT(Properties.cbBuffer <> 0);

  // Ask the allocator to reserve us some sample memory, NOTE the function
  // can succeed (that is return NOERROR) but still not have allocated the
  // memory that we requested, so we must check we got whatever we wanted

  Result := Alloc.SetProperties(Properties^, Actual);
  if FAILED(Result) then
    Exit;

  ASSERT(Actual.cBuffers = 1);

  if (Properties.cBuffers > Actual.cBuffers) or
     (Properties.cbBuffer > Actual.cbBuffer) then
    Result := E_FAIL
  else
    Result := NOERROR;
end;

var
  W: Integer = 768;
  H: Integer = 576;

function TEdgeDetectFilter.GetMediaType(Position: integer;
  out MediaType: pAM_Media_Type): HRESULT;
begin
  Result := E_FAIL;
  try
    // Is the input pin connected
    if not FInput.IsConnected then
    begin
      Result := E_UNEXPECTED;
      Exit;
    end;

    // This should never happen
    if (Position < 0) then
    begin
      Result := E_INVALIDARG;
      Exit;
    end;

    // Do we have more items to offer
    if (Position > 0) then
    begin
      Result := VFW_S_NO_MORE_ITEMS;
      Exit;
    end;

//  FInput.GetMediaType(0, MediaType);
    CopyMediaType(MediaType, FInput.CurrentMediaType.MediaType);

//  MediaType := FInput.CurrentMediaType.MediaType;

//  *pMediaType = m_pInput->CurrentMediaType();
    Result := S_OK;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.GetMediaType call...');
  end;
end;

function TEdgeDetectFilter.GetEdgeDetectType(
  out EdgeDetectType: Integer): HRESULT;
begin
  Result := E_FAIL;
  try
    EdgeDetectType := FEdgeDetectType;

    Result := S_OK;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.GetEdgeDetectType call...');
  end;
end;

function TEdgeDetectFilter.GetPages(out pages: TCAGUID): HRESULT;
begin
 Result := E_FAIL;
  try
    Pages.cElems := 1;
    Pages.pElems := CoTaskMemAlloc(SizeOf(TGUID));
    if (Pages.pElems = nil) then
    begin
      Result := E_OUTOFMEMORY;
      Exit;
    end;

    Pages.pElems^[0] := CLSID_JongovEdgeDetectPropertyPage;

    Result := S_OK;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.GetPages call...');
  end;
end;

function TEdgeDetectFilter.SetEdgeDetectType(
  EdgeDetectType: Integer): HRESULT;
begin
 Result := E_FAIL;
  try
    FCriticalSection.Enter;
    try
      FEdgeDetectType := EdgeDetectType;
    finally
      FCriticalSection.Leave;
    end;

    Result := S_OK;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.SetEdgeDetectType call...');
  end;
end;

function TEdgeDetectFilter.CurrCopyMediaSample(Source, Dest: IMediaSample): HRESULT;
var
  SourceBuffer, DestBuffer : PByte;
  SourceSize               : LongInt;
  TimeStart, TimeEnd       : TReference_Time;
  MediaStart, MediaEnd     : Int64;
  MediaType                : PAM_MEDIA_TYPE;
  EDT                      : Integer;
  K1_3, K2                 : Double;
begin
  Result := E_FAIL;
  try
    // Copy the sample data
    Source.GetPointer(SourceBuffer);
    SourceSize := Source.GetActualDataLength;

    Dest.GetPointer(DestBuffer);

    FCriticalSection.Enter;
    try
      EDT := FEdgeDetectType;
      K1_3 := FK1_3;
      K2 := FK2;
    finally
      FCriticalSection.Leave;
    end;

    if EDT = 1 then
    begin
      FEdgeDetector.HorizontalEdgeRGB32(SourceBuffer, FWidth, FHeight, K1_3, K2);
    end
    else if EDT = 2 then
    begin
      FEdgeDetector.VerticalEdgeRGB32(SourceBuffer, FWidth, FHeight, K1_3, K2);
    end
    else if EDT = 3 then
    begin
      FEdgeDetector.HorVerEdgeRGB32(SourceBuffer, FWidth, FHeight, K1_3, K2);
    end;

    CopyMemory(DestBuffer, SourceBuffer, SourceSize);

    Dest.SetActualDataLength(SourceSize);

    // Copy the sample times
    if (NOERROR = Source.GetTime(TimeStart, TimeEnd)) then
      Dest.SetTime(@TimeStart, @TimeEnd);

    if (Source.GetMediaTime(MediaStart,MediaEnd) = NOERROR) then
      Dest.SetMediaTime(@MediaStart, @MediaEnd);

    Dest.SetSyncPoint(Source.IsSyncPoint = S_OK);
    Dest.SetPreroll(Source.IsPreroll = S_OK);
    Dest.SetDiscontinuity(Source.IsDiscontinuity = S_OK);

    // Copy the media type
    MediaType := nil;
    if (Source.GetMediaType(MediaType) = S_OK) and (MediaType <> nil) then
    begin
      Dest.SetMediaType(MediaType^);
      DeleteMediaType(MediaType);
    end;

    Result := S_OK;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.CurrCopyMediaSample call...');
  end;
end;

function TEdgeDetectFilter.GetCoefficients(out K13, K2: Double): HRESULT;
begin
  Result := E_FAIL;
  try
    K13 := FK1_3;
    K2 := FK2;

    Result := S_OK;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.GetCoefficients call...');
  end;
end;

function TEdgeDetectFilter.SetCoefficients(K13, K2: Double): HRESULT;
begin
  Result := E_FAIL;
  try
    FCriticalSection.Enter;
    try
      FK1_3 := K13;
      FK2 := K2;
    finally
      FCriticalSection.Leave;
    end;

    Result := S_OK;
  except
    OutputDebugString('tsv: Error in TEdgeDetectFilter.SetCoefficients call...');
  end;
end;

initialization
  TBCClassFactory.CreateFilter(TEdgeDetectFilter, 'Jongov Edge Detect Filter', CLSID_JongovEdgeDetectFilter,
                               CLSID_LegacyAmFilterCategory, MERIT_DO_NOT_USE, 0, nil);
end.
